from logging import getLogger

from qbittorrentapi.decorators import login_required
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.exceptions import LoginFailed
from qbittorrentapi.request import Request

logger = getLogger(__name__)


class Authorization(ClientCache):
    """
    Allows interaction with the "Authorization" API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> client.auth.is_logged_in
        >>> client.auth.log_in(username='admin', password='adminadmin')
        >>> client.auth.log_out()
    """

    @property
    def is_logged_in(self):
        """Implements :meth:`~AuthAPIMixIn.is_logged_in`"""
        return self._client.is_logged_in

    def log_in(self, username=None, password=None):
        """Implements :meth:`~AuthAPIMixIn.auth_log_in`"""
        return self._client.auth_log_in(username=username, password=password)

    def log_out(self):
        """Implements :meth:`~AuthAPIMixIn.auth_log_out`"""
        return self._client.auth_log_out()


class AuthAPIMixIn(Request):
    """
    Implementation of all Authorization API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> client.auth_is_logged_in()
        >>> client.auth_log_in(username='admin', password='adminadmin')
        >>> client.auth_log_out()
    """

    @property
    def auth(self):
        """
        Allows for transparent interaction with Authorization endpoints.

        :return: Auth object
        """
        if self._authorization is None:
            self._authorization = Authorization(client=self)
        return self._authorization

    authorization = auth

    @property
    def is_logged_in(self):
        """
        Returns True/False for whether a log-in attempt was ever successfully completed.

        It isn't possible to know if qBittorrent will accept whatever SID is locally
        cached...however, any request that is rejected because of the SID will be automatically
        retried after a new SID is requested.

        :returns: True/False for whether a log-in attempt was previously completed
        """
        return bool(self._SID)

    def auth_log_in(self, username=None, password=None, **kwargs):
        """
        Log in to qBittorrent host.

        :raises LoginFailed: if credentials failed to log in
        :raises Forbidden403Error: if user user is banned...or not logged in

        :param username: user name for qBittorrent client
        :param password: password for qBittorrent client
        :return: None
        """
        if username:
            self.username = username or ""
            self._password = password or ""

        self._initialize_context()
        self._post(
            _name=APINames.Authorization,
            _method="login",
            data={"username": self.username, "password": self._password},
            **kwargs
        )

        if not self.is_logged_in:
            logger.debug('Login failed for user "%s"' % self.username)
            raise self._suppress_context(
                LoginFailed('Login authorization failed for user "%s"' % self.username)
            )
        else:
            logger.debug('Login successful for user "%s"' % self.username)
            logger.debug("SID: %s" % self._SID)

    @property
    def _SID(self):
        """
        Authorization cookie from qBittorrent.

        :return: SID auth cookie from qBittorrent or None if one isn't already acquired
        """
        if self._requests_session:
            return self._requests_session.cookies.get("SID", None)
        return None

    @login_required
    def auth_log_out(self, **kwargs):
        """
        End session with qBittorrent.
        """
        self._get(_name=APINames.Authorization, _method="logout", **kwargs)



from qbittorrentapi.app import AppAPIMixIn
from qbittorrentapi.auth import AuthAPIMixIn
from qbittorrentapi.log import LogAPIMixIn
from qbittorrentapi.sync import SyncAPIMixIn
from qbittorrentapi.transfer import TransferAPIMixIn
from qbittorrentapi.torrents import TorrentsAPIMixIn
from qbittorrentapi.rss import RSSAPIMixIn
from qbittorrentapi.search import SearchAPIMixIn

# NOTES
# Implementation
#     Required API parameters
#         - To avoid runtime errors, required API parameters are not explicitly
#           enforced in the code. Instead, I found if qBittorrent returns HTTP400
#           without am error message, at least one required parameter is missing.
#           This raises a MissingRequiredParameters400 error.
#         - Alternatively, if a parameter is malformatted, HTTP400 is returned
#           with an error message.
#           This raises a InvalidRequest400 error.
#
#     Unauthorized HTTP 401
#         - This is only raised if XSS is detected or host header validation fails.
#
# API Peculiarities
#     app/setPreferences
#         - This was endlessly frustrating since it requires data in the
#           form of {'json': dumps({'dht': True})}...this way, Requests sends the
#           JSON dump as a key/value pair for "json" via x-www-form-urlencoded.
#         - Sending an empty string for 'banned_ips' drops the useless message
#           below in to the log file (same for WebUI):
#             ' is not a valid IP address and was rejected while applying the list of banned addresses.'
#             - [Resolved] https://github.com/qbittorrent/qBittorrent/issues/10745
#
#     torrents/downloadLimit and uploadLimit
#         - Hashes handling is non-standard. 404 is not returned for bad hashes and 'all' doesn't work.
#         - https://github.com/qbittorrent/qBittorrent/blob/6de02b0f2a79eeb4d7fb624c39a9f65ffe181d68/src/webui/api/torrentscontroller.cpp#L754
#         - https://github.com/qbittorrent/qBittorrent/issues/10744
#
#     torrents/info
#         - when using a GET request, the params (such as category) seemingly can't
#           contain spaces; however, POST does work with spaces.
#         - [Resolved] https://github.com/qbittorrent/qBittorrent/issues/10606


class Client(
    AppAPIMixIn,
    AuthAPIMixIn,
    LogAPIMixIn,
    SyncAPIMixIn,
    TransferAPIMixIn,
    TorrentsAPIMixIn,
    RSSAPIMixIn,
    SearchAPIMixIn,
):

    """
    Initialize API for qBittorrent client.

    Host must be specified. Username and password can be specified at login.
    A call to :meth:`~qbittorrentapi.auth.AuthAPIMixIn.auth_log_in` is not explicitly
    required if username and password are provided during Client construction.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> torrents = client.torrents_info()

    :param host: hostname for qBittorrent Web API (e.g. [http[s]://]localhost[:8080])
    :param port: port number for qBittorrent Web API (note: only used if host does not contain a port)
    :param username: user name for qBittorrent client
    :param password: password for qBittorrent client

    :param SIMPLE_RESPONSES: By default, complex objects are returned from some endpoints. These objects will allow for
        accessing responses' items as attributes and include methods for contextually relevant actions.
        This comes at the cost of performance. Generally, this cost isn't large; however, some
        endpoints, such as torrents_files() method, may need to convert a large payload.
        Set this to True to return the simple JSON back.
        Alternatively, set this to True only for an individual method call. For instance, when
        requesting the files for a torrent: client.torrents_files(hash='...', SIMPLE_RESPONSES=True).
    :param VERIFY_WEBUI_CERTIFICATE: Set to False to skip verify certificate for HTTPS connections;
        for instance, if the connection is using a self-signed certificate. Not setting this to False for self-signed
        certs will cause a APIConnectionError exception to be raised.
    :param RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS: Some Endpoints may not be implemented in older
        versions of qBittorrent. Setting this to True will raise a NotImplementedError instead of just returning None.
    :param DISABLE_LOGGING_DEBUG_OUTPUT: Turn off debug output from logging for this package as well as Requests & urllib3.
    """

    def __init__(self, host="", port=None, username=None, password=None, **kwargs):
        super(Client, self).__init__(
            host=host, port=port, username=username, password=password, **kwargs
        )



from qbittorrentapi.decorators import Alias
from qbittorrentapi.decorators import aliased
from qbittorrentapi.decorators import handle_hashes
from qbittorrentapi.decorators import login_required
from qbittorrentapi.decorators import response_json
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.request import Request


class SyncMainDataDictionary(Dictionary):
    """Response for :meth:`~SyncAPIMixIn.sync_maindata`"""


class SyncTorrentPeersDictionary(Dictionary):
    """Response for :meth:`~SyncAPIMixIn.sync_torrent_peers`"""


class Sync(ClientCache):
    """
    Alows interaction with the "Sync" API endpoints.

    Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> # this are all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'sync_' prepended)
        >>> maindata = client.sync.maindata(rid="...")
        >>> # for use when continuously calling maindata for changes in torrents
        >>> # this will automatically request the changes since the last call
        >>> md = client.sync.maindata.delta()
        >>> #
        >>> torrentPeers = client.sync.torrentPeers(hash="...'", rid='...')
        >>> torrent_peers = client.sync.torrent_peers(hash="...'", rid='...')
    """

    def __init__(self, client):
        super(Sync, self).__init__(client=client)
        self.maindata = self._MainData(client=client)
        self.torrent_peers = self._TorrentPeers(client=client)
        self.torrentPeers = self.torrent_peers

    class _MainData(ClientCache):
        def __init__(self, client):
            super(Sync._MainData, self).__init__(client=client)
            self._rid = 0

        def __call__(self, rid=None, **kwargs):
            return self._client.sync_maindata(rid=rid, **kwargs)

        def delta(self, **kwargs):
            md = self._client.sync_maindata(rid=self._rid, **kwargs)
            self._rid = md.get("rid", 0)
            return md

        def reset_rid(self):
            self._rid = 0

    class _TorrentPeers(ClientCache):
        def __init__(self, client):
            super(Sync._TorrentPeers, self).__init__(client=client)
            self._rid = None

        def __call__(self, torrent_hash=None, rid=None, **kwargs):
            return self._client.sync_torrent_peers(
                torrent_hash=torrent_hash, rid=rid, **kwargs
            )

        def delta(self, torrent_hash=None, **kwargs):
            tp = self._client.sync_torrent_peers(
                torrent_hash=torrent_hash, rid=self._rid, **kwargs
            )
            self._rid = tp.get("rid", 0)
            return tp

        def reset_rid(self):
            self._rid = 0


@aliased
class SyncAPIMixIn(Request):
    """
    Implementation of all Sync API Methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> maindata = client.sync_maindata(rid="...")
        >>> torrent_peers = client.sync_torrent_peers(torrent_hash="...'", rid='...')
    """

    @property
    def sync(self):
        """
        Allows for transparent interaction with Sync endpoints.

        See Sync class for usage.
        :return: Transfer object
        """
        if self._sync is None:
            self._sync = Sync(client=self)
        return self._sync

    @response_json(SyncMainDataDictionary)
    @login_required
    def sync_maindata(self, rid=0, **kwargs):
        """
        Retrieves sync data.

        :param rid: response ID
        :return: :class:`SyncMainDataDictionary` - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-main-data
        """
        data = {"rid": rid}
        return self._post(_name=APINames.Sync, _method="maindata", data=data, **kwargs)

    @Alias("sync_torrentPeers")
    @handle_hashes
    @response_json(SyncTorrentPeersDictionary)
    @login_required
    def sync_torrent_peers(self, torrent_hash=None, rid=0, **kwargs):
        """
        Retrieves torrent sync data. (alias: sync_torrentPeers)

        :raises NotFound404Error:

        :param torrent_hash: hash for torrent
        :param rid: response ID
        :return: :class:`SyncTorrentPeersDictionary` - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-torrent-peers-data
        """
        data = {"hash": torrent_hash, "rid": rid}
        return self._post(
            _name=APINames.Sync, _method="torrentPeers", data=data, **kwargs
        )



from qbittorrentapi.decorators import Alias
from qbittorrentapi.decorators import aliased
from qbittorrentapi.decorators import login_required
from qbittorrentapi.decorators import response_json
from qbittorrentapi.decorators import response_text
from qbittorrentapi.decorators import endpoint_introduced
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.request import Request


class TransferInfoDictionary(Dictionary):
    """Response to :meth:`~TransferAPIMixIn.transfer_info`"""


@aliased
class Transfer(ClientCache):
    """
    Alows interaction with the "Transfer" API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> # this are all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'transfer_' prepended)
        >>> transfer_info = client.transfer.info
        >>> # access and set download/upload limits as attributes
        >>> dl_limit = client.transfer.download_limit
        >>> # this updates qBittorrent in real-time
        >>> client.transfer.download_limit = 1024000
        >>> # update speed limits mode to alternate or not
        >>> client.transfer.speedLimitsMode = True
    """

    @property
    def info(self):
        """Implements :meth:`~TransferAPIMixIn.transfer_info`"""
        return self._client.transfer_info()

    @property
    def speed_limits_mode(self):
        """Implements :meth:`~TransferAPIMixIn.transfer_speed_limits_mode`"""
        return self._client.transfer_speed_limits_mode()

    speedLimitsMode = speed_limits_mode

    @speedLimitsMode.setter
    def speedLimitsMode(self, v):
        """Implements :meth:`~TransferAPIMixIn.transfer_toggle_speed_limits_mode`"""
        self.speed_limits_mode = v

    @speed_limits_mode.setter
    def speed_limits_mode(self, v):
        """Implements :meth:`~TransferAPIMixIn.transfer_toggle_speed_limits_mode`"""
        self.toggle_speed_limits_mode(intended_state=v)

    @Alias("toggleSpeedLimitsMode")
    def toggle_speed_limits_mode(self, intended_state=None, **kwargs):
        """Implements :meth:`~TransferAPIMixIn.transfer_toggle_speed_limits_mode`"""
        return self._client.transfer_toggle_speed_limits_mode(
            intended_state=intended_state, **kwargs
        )

    @property
    def download_limit(self):
        """Implements :meth:`~TransferAPIMixIn.transfer_download_limit`"""
        return self._client.transfer_download_limit()

    downloadLimit = download_limit

    @downloadLimit.setter
    def downloadLimit(self, v):
        """Implements :meth:`~TransferAPIMixIn.transfer_set_download_limit`"""
        self.download_limit = v

    @download_limit.setter
    def download_limit(self, v):
        """Implements :meth:`~TransferAPIMixIn.transfer_set_download_limit`"""
        self.set_download_limit(limit=v)

    @property
    def upload_limit(self):
        """Implements :meth:`~TransferAPIMixIn.transfer_upload_limit`"""
        return self._client.transfer_upload_limit()

    uploadLimit = upload_limit

    @uploadLimit.setter
    def uploadLimit(self, v):
        """Implements :meth:`~TransferAPIMixIn.transfer_set_upload_limit`"""
        self.upload_limit = v

    @upload_limit.setter
    def upload_limit(self, v):
        """Implements :meth:`~TransferAPIMixIn.transfer_set_upload_limit`"""
        self.set_upload_limit(limit=v)

    @Alias("setDownloadLimit")
    def set_download_limit(self, limit=None, **kwargs):
        """Implements :meth:`~TransferAPIMixIn.transfer_set_download_limit`"""
        return self._client.transfer_set_download_limit(limit=limit, **kwargs)

    @Alias("setUploadLimit")
    def set_upload_limit(self, limit=None, **kwargs):
        """Implements :meth:`~TransferAPIMixIn.transfer_set_upload_limit`"""
        return self._client.transfer_set_upload_limit(limit=limit, **kwargs)

    @Alias("banPeers")
    def ban_peers(self, peers=None, **kwargs):
        """Implements :meth:`~TransferAPIMixIn.transfer_ban_peers`"""
        return self._client.transfer_ban_peers(peers=peers, **kwargs)


@aliased
class TransferAPIMixIn(Request):
    """
    Implementation of all Transfer API methods

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> transfer_info = client.transfer_info()
        >>> client.transfer_set_download_limit(limit=1024000)
    """

    @property
    def transfer(self):
        """
        Allows for transparent interaction with Transfer endpoints.

        See Transfer class for usage.
        :return: Transfer object
        """
        if self._transfer is None:
            self._transfer = Transfer(client=self)
        return self._transfer

    @response_json(TransferInfoDictionary)
    @login_required
    def transfer_info(self, **kwargs):
        """
        Retrieves the global transfer info usually found in qBittorrent status bar.

        :return: :class:`TransferInfoDictionary` - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-global-transfer-info
        """
        return self._get(_name=APINames.Transfer, _method="info", **kwargs)

    @Alias("transfer_speedLimitsMode")
    @response_text(str)
    @login_required
    def transfer_speed_limits_mode(self, **kwargs):
        """
        Retrieves whether alternative speed limits are enabled. (alias: transfer_speedLimitMode)

        :return: '1' if alternative speed limits are currently enabled, '0' otherwise
        """
        return self._get(_name=APINames.Transfer, _method="speedLimitsMode", **kwargs)

    @Alias("transfer_toggleSpeedLimitsMode")
    @login_required
    def transfer_toggle_speed_limits_mode(self, intended_state=None, **kwargs):
        """
        Sets whether alternative speed limits are enabled. (alias: transfer_toggleSpeedLimitsMode)

        :param intended_state: True to enable alt speed and False to disable.
                               Leaving None will toggle the current state.
        :return: None
        """
        is_enabled = lambda: self.transfer_speed_limits_mode() == "1"
        if intended_state is None or is_enabled() is not intended_state:
            self._post(
                _name=APINames.Transfer, _method="toggleSpeedLimitsMode", **kwargs
            )

    @Alias("transfer_downloadLimit")
    @response_text(int)
    @login_required
    def transfer_download_limit(self, **kwargs):
        """
        Retrieves download limit. 0 is unlimited. (alias: transfer_downloadLimit)

        :return: integer
        """
        return self._get(_name=APINames.Transfer, _method="downloadLimit", **kwargs)

    @Alias("transfer_uploadLimit")
    @response_text(int)
    @login_required
    def transfer_upload_limit(self, **kwargs):
        """
        Retrieves upload limit. 0 is unlimited. (alias: transfer_uploadLimit)

        :return: integer
        """
        return self._get(_name=APINames.Transfer, _method="uploadLimit", **kwargs)

    @Alias("transfer_setDownloadLimit")
    @login_required
    def transfer_set_download_limit(self, limit=None, **kwargs):
        """
        Set the global download limit in bytes/second. (alias: transfer_setDownloadLimit)

        :param limit: download limit in bytes/second (0 or -1 for no limit)
        :return: None
        """
        data = {"limit": limit}
        self._post(
            _name=APINames.Transfer, _method="setDownloadLimit", data=data, **kwargs
        )

    @Alias("transfer_setUploadLimit")
    @login_required
    def transfer_set_upload_limit(self, limit=None, **kwargs):
        """
        Set the global download limit in bytes/second. (alias: transfer_setUploadLimit)

        :param limit: upload limit in bytes/second (0 or -1 for no limit)
        :return: None
        """
        data = {"limit": limit}
        self._post(
            _name=APINames.Transfer, _method="setUploadLimit", data=data, **kwargs
        )

    @Alias("transfer_banPeers")
    @endpoint_introduced("2.3", "transfer/banPeers")
    @login_required
    def transfer_ban_peers(self, peers=None, **kwargs):
        """
        Ban one or more peers. (alias: transfer_banPeers)

        :param peers: one or more peers to ban. each peer should take the form 'host:port'
        :return: None
        """
        data = {"peers": self._list2string(peers, "|")}
        self._post(_name=APINames.Transfer, _method="banPeers", data=data, **kwargs)



from qbittorrentapi.decorators import Alias
from qbittorrentapi.decorators import aliased
from qbittorrentapi.decorators import endpoint_introduced
from qbittorrentapi.decorators import login_required
from qbittorrentapi.decorators import response_json
from qbittorrentapi.decorators import version_removed
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.definitions import List
from qbittorrentapi.definitions import ListEntry
from qbittorrentapi.request import Request


class SearchJobDictionary(Dictionary):
    """Response for :meth:`~SearchAPIMixIn.search_start`"""

    def __init__(self, data, client):
        self._search_job_id = data.get("id", None)
        super(SearchJobDictionary, self).__init__(data=data, client=client)

    def stop(self, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_stop`"""
        return self._client.search_stop(search_id=self._search_job_id, **kwargs)

    def status(self, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_status`"""
        return self._client.search_status(search_id=self._search_job_id, **kwargs)

    def results(self, limit=None, offset=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_results`"""
        return self._client.search_results(
            limit=limit, offset=offset, search_id=self._search_job_id, **kwargs
        )

    def delete(self, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_delete`"""
        return self._client.search_delete(search_id=self._search_job_id, **kwargs)


class SearchResultsDictionary(Dictionary):
    """Response for :meth:`~SearchAPIMixIn.search_results`"""


class SearchStatusesList(List):
    """Response for :meth:`~SearchAPIMixIn.search_status`"""

    def __init__(self, list_entries=None, client=None):
        super(SearchStatusesList, self).__init__(
            list_entries, entry_class=SearchStatus, client=client
        )


class SearchStatus(ListEntry):
    """Item in :class:`SearchStatusesList`"""


class SearchCategoriesList(List):
    """Response for :meth:`~SearchAPIMixIn.search_categories`"""

    def __init__(self, list_entries=None, client=None):
        super(SearchCategoriesList, self).__init__(
            list_entries, entry_class=SearchCategory, client=client
        )


class SearchCategory(ListEntry):
    """Item in :class:`SearchCategoriesList`"""


class SearchPluginsList(List):
    """Response for :meth:`~SearchAPIMixIn.search_plugins`"""

    def __init__(self, list_entries=None, client=None):
        super(SearchPluginsList, self).__init__(
            list_entries, entry_class=SearchPlugin, client=client
        )


class SearchPlugin(ListEntry):
    """Item in :class:`SearchPluginsList`"""


@aliased
class Search(ClientCache):
    """
    Allows interaction with "Search" API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'search_' prepended)
        >>> # initiate searches and retrieve results
        >>> search_job = client.search.start(pattern='Ubuntu', plugins='all', category='all')
        >>> status = search_job.status()
        >>> results = search_job.result()
        >>> search_job.delete()
        >>> # inspect and manage plugins
        >>> plugins = client.search.plugins
        >>> cats = client.search.categories(plugin_name='...')
        >>> client.search.install_plugin(sources='...')
        >>> client.search.update_plugins()
    """

    def start(self, pattern=None, plugins=None, category=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_start`"""
        return self._client.search_start(
            pattern=pattern, plugins=plugins, category=category, **kwargs
        )

    def stop(self, search_id=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_stop`"""
        return self._client.search_stop(search_id=search_id, **kwargs)

    def status(self, search_id=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_status`"""
        return self._client.search_status(search_id=search_id, **kwargs)

    def results(self, search_id=None, limit=None, offset=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_results`"""
        return self._client.search_results(
            search_id=search_id, limit=limit, offset=offset, **kwargs
        )

    def delete(self, search_id=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_delete`"""
        return self._client.search_delete(search_id=search_id, **kwargs)

    def categories(self, plugin_name=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_categories`"""
        return self._client.search_categories(plugin_name=plugin_name, **kwargs)

    @property
    def plugins(self):
        """Implements :meth:`~SearchAPIMixIn.search_plugins`"""
        return self._client.search_plugins()

    @Alias("installPlugin")
    def install_plugin(self, sources=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_install_plugin`"""
        return self._client.search_install_plugin(sources=sources, **kwargs)

    @Alias("uninstallPlugin")
    def uninstall_plugin(self, sources=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_uninstall_plugin`"""
        return self._client.search_uninstall_plugin(sources=sources, **kwargs)

    @Alias("enablePlugin")
    def enable_plugin(self, plugins=None, enable=None, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_enable_plugin`"""
        return self._client.search_enable_plugin(
            plugins=plugins, enable=enable, **kwargs
        )

    @Alias("updatePlugins")
    def update_plugins(self, **kwargs):
        """Implements :meth:`~SearchAPIMixIn.search_update_plugins`"""
        return self._client.search_update_plugins(**kwargs)


@aliased
class SearchAPIMixIn(Request):
    """
    Implementation for all Search API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> search_job = client.search_start(pattern='Ubuntu', plugins='all', category='all')
        >>> client.search_stop(search_id=search_job.id)
        >>> # or
        >>> search_job.stop()
        >>>
    """

    @property
    def search(self):
        """
        Allows for transparent interaction with Search endpoints.

        See Search class for usage.
        :return: Search object
        """
        if self._search is None:
            self._search = Search(client=self)
        return self._search

    @endpoint_introduced("2.1.1", "search/start")
    @response_json(SearchJobDictionary)
    @login_required
    def search_start(self, pattern=None, plugins=None, category=None, **kwargs):
        """
        Start a search. Python must be installed. Host may limit number of concurrent searches.

        :raises Conflict409Error:

        :param pattern: term to search for
        :param plugins: list of plugins to use for searching (supports 'all' and 'enabled')
        :param category: categories to limit search; dependent on plugins. (supports 'all')
        :return: :class:`SearchJobDictionary`
        """
        data = {
            "pattern": pattern,
            "plugins": self._list2string(plugins, "|"),
            "category": category,
        }
        return self._post(_name=APINames.Search, _method="start", data=data, **kwargs)

    @endpoint_introduced("2.1.1", "search/stop")
    @login_required
    def search_stop(self, search_id=None, **kwargs):
        """
        Stop a running search.

        :raises NotFound404Error:

        :param search_id: ID of search job to stop
        :return: None
        """
        data = {"id": search_id}
        self._post(_name=APINames.Search, _method="stop", data=data, **kwargs)

    @endpoint_introduced("2.1.1", "search/status")
    @response_json(SearchStatusesList)
    @login_required
    def search_status(self, search_id=None, **kwargs):
        """
        Retrieve status of one or all searches.

        :raises NotFound404Error:

        :param search_id: ID of search to get status; leave emtpy for status of all jobs
        :return: :class:`SearchStatusesList` - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-search-status
        """
        params = {"id": search_id}
        return self._get(
            _name=APINames.Search, _method="status", params=params, **kwargs
        )

    @endpoint_introduced("2.1.1", "search/results")
    @response_json(SearchResultsDictionary)
    @login_required
    def search_results(self, search_id=None, limit=None, offset=None, **kwargs):
        """
        Retrieve the results for the search.

        :raises NotFound404Error:
        :raises Conflict409Error:

        :param search_id: ID of search job
        :param limit: number of results to return
        :param offset: where to start returning results
        :return: :class:`SearchResultsDictionary` - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-search-results
        """
        data = {"id": search_id, "limit": limit, "offset": offset}
        return self._post(_name=APINames.Search, _method="results", data=data, **kwargs)

    @endpoint_introduced("2.1.1", "search/delete")
    @login_required
    def search_delete(self, search_id=None, **kwargs):
        """
        Delete a search job.

        :raises NotFound404Error:

        :param search_id: ID of search to delete
        :return: None
        """
        data = {"id": search_id}
        self._post(_name=APINames.Search, _method="delete", data=data, **kwargs)

    @endpoint_introduced("2.1.1", "search/categories")
    @version_removed("2.6", "search/categories")
    @response_json(SearchCategoriesList)
    @login_required
    def search_categories(self, plugin_name=None, **kwargs):
        """
        Retrieve categories for search.
        Note: endpoint was removed in qBittorrent v4.3.0

        :param plugin_name: Limit categories returned by plugin(s) (supports 'all' and 'enabled')
        :return: :class:`SearchCategoriesList`
        """
        data = {"pluginName": plugin_name}
        return self._post(
            _name=APINames.Search, _method="categories", data=data, **kwargs
        )

    @endpoint_introduced("2.1.1", "search/plugins")
    @response_json(SearchPluginsList)
    @login_required
    def search_plugins(self, **kwargs):
        """
        Retrieve details of search plugins.

        :return: :class:`SearchPluginsList` - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-search-plugins
        """
        return self._get(_name=APINames.Search, _method="plugins", **kwargs)

    @endpoint_introduced("2.1.1", "search/installPlugin")
    @Alias("search_installPlugin")
    @login_required
    def search_install_plugin(self, sources=None, **kwargs):
        """
        Install search plugins from either URL or file. (alias: search_installPlugin)

        :param sources: list of URLs or filepaths
        :return: None
        """
        data = {"sources": self._list2string(sources, "|")}
        self._post(_name=APINames.Search, _method="installPlugin", data=data, **kwargs)

    @endpoint_introduced("2.1.1", "search/uninstallPlugin")
    @Alias("search_uninstallPlugin")
    @login_required
    def search_uninstall_plugin(self, names=None, **kwargs):
        """
        Uninstall search plugins. (alias: search_uninstallPlugin)

        :param names: names of plugins to uninstall
        :return: None
        """
        data = {"names": self._list2string(names, "|")}
        self._post(
            _name=APINames.Search, _method="uninstallPlugin", data=data, **kwargs
        )

    @endpoint_introduced("2.1.1", "search/enablePlugin")
    @Alias("search_enablePlugin")
    @login_required
    def search_enable_plugin(self, plugins=None, enable=None, **kwargs):
        """
        Enable or disable search plugin(s). (alias: search_enablePlugin)

        :param plugins: list of plugin names
        :param enable: True or False
        :return: None
        """
        data = {"names": self._list2string(plugins, "|"), "enable": enable}
        self._post(_name=APINames.Search, _method="enablePlugin", data=data, **kwargs)

    @endpoint_introduced("2.1.1", "search/updatePlugin")
    @Alias("search_updatePlugins")
    @login_required
    def search_update_plugins(self, **kwargs):
        """
        Auto update search plugins. (alias: search_updatePlugins)

        :return: None
        """
        self._post(_name=APINames.Search, _method="updatePlugins", **kwargs)



from qbittorrentapi.decorators import login_required
from qbittorrentapi.decorators import response_json
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import List, ListEntry
from qbittorrentapi.request import Request


class LogPeersList(List):
    """Response for :meth:`~LogAPIMixIn.log_peers`"""

    def __init__(self, list_entries=None, client=None):
        super(LogPeersList, self).__init__(
            list_entries, entry_class=LogPeer, client=client
        )


class LogPeer(ListEntry):
    """Item in :class:`LogPeersList`"""


class LogMainList(List):
    """Response to :meth:`~LogAPIMixIn.log_main`"""

    def __init__(self, list_entries=None, client=None):
        super(LogMainList, self).__init__(
            list_entries, entry_class=LogEntry, client=client
        )


class LogEntry(ListEntry):
    """Item in :class:`LogMainList`"""


class Log(ClientCache):
    """
    Allows interaction with "Log" API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'log_' prepended)
        >>> log_list = client.log.main()
        >>> peers_list = client.log.peers(hash='...')
        >>> # can also filter log down with additional attributes
        >>> log_info = client.log.main.info(last_known_id='...')
        >>> log_warning = client.log.main.warning(last_known_id='...')
    """

    def __init__(self, client):
        super(Log, self).__init__(client=client)
        self.main = Log._Main(client=client)

    def peers(self, last_known_id=None, **kwargs):
        """Implements :meth:`~LogAPIMixIn.log_peers`"""
        return self._client.log_peers(last_known_id=last_known_id, **kwargs)

    class _Main(ClientCache):
        def _api_call(
            self,
            normal=None,
            info=None,
            warning=None,
            critical=None,
            last_known_id=None,
            **kwargs
        ):
            return self._client.log_main(
                normal=normal,
                info=info,
                warning=warning,
                critical=critical,
                last_known_id=last_known_id,
                **kwargs
            )

        def __call__(
            self,
            normal=None,
            info=None,
            warning=None,
            critical=None,
            last_known_id=None,
            **kwargs
        ):
            return self._api_call(
                normal=normal,
                info=info,
                warning=warning,
                critial=critical,
                last_known_id=last_known_id,
                **kwargs
            )

        def info(self, last_known_id=None, **kwargs):
            return self._api_call(last_known_id=last_known_id, **kwargs)

        def normal(self, last_known_id=None, **kwargs):
            return self._api_call(info=False, last_known_id=last_known_id, **kwargs)

        def warning(self, last_known_id=None, **kwargs):
            return self._api_call(
                info=False, normal=False, last_known_id=last_known_id, **kwargs
            )

        def critical(self, last_known_id=None, **kwargs):
            return self._api_call(
                info=False,
                normal=False,
                warning=False,
                last_known_id=last_known_id,
                **kwargs
            )


class LogAPIMixIn(Request):
    """
    Implementation of all Log API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> client.log_main(info=False)
        >>> client.log_peers()
    """

    @property
    def log(self):
        """
        Allows for transparent interaction with Log endpoints.

        See Log class for usage.
        :return: Log object
        """
        if self._log is None:
            self._log = Log(client=self)
        return self._log

    @response_json(LogMainList)
    @login_required
    def log_main(
        self,
        normal=None,
        info=None,
        warning=None,
        critical=None,
        last_known_id=None,
        **kwargs
    ):
        """
        Retrieve the qBittorrent log entries. Iterate over returned object.

        :param normal: False to exclude 'normal' entries
        :param info: False to exclude 'info' entries
        :param warning: False to exclude 'warning' entries
        :param critical: False to exclude 'critical' entries
        :param last_known_id: only entries with an ID greater than this value will be returned
        :return: :class:`LogMainList`
        """
        parameters = {
            "normal": normal,
            "info": info,
            "warning": warning,
            "critical": critical,
            "last_known_id": last_known_id,
        }
        return self._get(
            _name=APINames.Log, _method="main", params=parameters, **kwargs
        )

    @response_json(LogPeersList)
    @login_required
    def log_peers(self, last_known_id=None, **kwargs):
        """
        Retrieve qBittorrent peer log.

        :param last_known_id: only entries with an ID greater than this value will be returned
        :return: :class:`LogPeersList`
        """
        parameters = {"last_known_id": last_known_id}
        return self._get(
            _name=APINames.Log, _method="peers", params=parameters, **kwargs
        )



from json import dumps

from qbittorrentapi.decorators import Alias
from qbittorrentapi.decorators import aliased
from qbittorrentapi.decorators import endpoint_introduced
from qbittorrentapi.decorators import login_required
from qbittorrentapi.decorators import response_json
from qbittorrentapi.definitions import APINames
from qbittorrentapi.definitions import ClientCache
from qbittorrentapi.definitions import Dictionary
from qbittorrentapi.request import Request


class RSSitemsDictionary(Dictionary):
    """Response for :meth:`~RSSAPIMixIn.rss_items`"""


class RSSRulesDictionary(Dictionary):
    """Response for :meth:`~RSSAPIMixIn.rss_rules`"""


@aliased
class RSS(ClientCache):
    """
    Allows interaction with "RSS" API endpoints.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> # this is all the same attributes that are available as named in the
        >>> #  endpoints or the more pythonic names in Client (with or without 'log_' prepended)
        >>> rss_rules = client.rss.rules
        >>> client.rss.addFolder(folder_path="TPB")
        >>> client.rss.addFeed(url='...', item_path="TPB\\Top100")
        >>> client.rss.remove_item(item_path="TPB") # deletes TPB and Top100
        >>> client.rss.set_rule(rule_name="...", rule_def={...})
        >>> items = client.rss.items.with_data
        >>> items = client.rss.items.without_data
    """

    def __init__(self, client):
        super(RSS, self).__init__(client=client)
        self.items = RSS._Items(client=client)

    @Alias("addFolder")
    def add_folder(self, folder_path=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_add_folder`"""
        return self._client.rss_add_folder(folder_path=folder_path, **kwargs)

    @Alias("addFeed")
    def add_feed(self, url=None, item_path=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_add_feed`"""
        return self._client.rss_add_feed(url=url, item_path=item_path, **kwargs)

    @Alias("removeItem")
    def remove_item(self, item_path=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_remove_item`"""
        return self._client.rss_remove_item(item_path=item_path, **kwargs)

    @Alias("moveItem")
    def move_item(self, orig_item_path=None, new_item_path=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_move_item`"""
        return self._client.rss_move_item(
            orig_item_path=orig_item_path, new_item_path=new_item_path, **kwargs
        )

    @Alias("refreshItem")
    def refresh_item(self, item_path=None):
        """Implements :meth:`~RSSAPIMixIn.rss_refresh_item`"""
        return self._client.rss_refresh_item(item_path=item_path)

    @Alias("markAsRead")
    def mark_as_read(self, item_path=None, article_id=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_mark_as_read`"""
        return self._client.rss_mark_as_read(
            item_path=item_path, article_id=article_id, **kwargs
        )

    @Alias("setRule")
    def set_rule(self, rule_name=None, rule_def=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_set_rule`"""
        return self._client.rss_set_rule(
            rule_name=rule_name, rule_def=rule_def, **kwargs
        )

    @Alias("renameRule")
    def rename_rule(self, orig_rule_name=None, new_rule_name=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_rename_rule`"""
        return self._client.rss_rename_rule(
            orig_rule_name=orig_rule_name, new_rule_name=new_rule_name, **kwargs
        )

    @Alias("removeRule")
    def remove_rule(self, rule_name=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_remove_rule`"""
        return self._client.rss_remove_rule(rule_name=rule_name, **kwargs)

    @property
    def rules(self):
        """Implements :meth:`~RSSAPIMixIn.rss_rules`"""
        return self._client.rss_rules()

    @Alias("matchingArticles")
    def matching_articles(self, rule_name=None, **kwargs):
        """Implements :meth:`~RSSAPIMixIn.rss_matching_articles`"""
        return self._client.rss_matching_articles(rule_name=rule_name, **kwargs)

    class _Items(ClientCache):
        def __call__(self, include_feed_data=None, **kwargs):
            return self._client.rss_items(include_feed_data=include_feed_data, **kwargs)

        @property
        def without_data(self):
            return self._client.rss_items(include_feed_data=False)

        @property
        def with_data(self):
            return self._client.rss_items(include_feed_data=True)


@aliased
class RSSAPIMixIn(Request):
    """
    Implementation of all RSS API methods.

    :Usage:
        >>> from qbittorrentapi import Client
        >>> client = Client(host='localhost:8080', username='admin', password='adminadmin')
        >>> rss_rules = client.rss_rules()
        >>> client.rss_set_rule(rule_name="...", rule_def={...})
    """

    @property
    def rss(self):
        """
        Allows for transparent interaction with RSS endpoints.

        See RSS class for usage.
        :return: RSS object
        """
        if self._rss is None:
            self._rss = RSS(client=self)
        return self._rss

    @Alias("rss_addFolder")
    @login_required
    def rss_add_folder(self, folder_path=None, **kwargs):
        """
        Add a RSS folder. Any intermediate folders in path must already exist. (alias: rss_addFolder)

        :raises Conflict409Error:

        :param folder_path: path to new folder (e.g. Linux\ISOs)
        :return: None
        """
        data = {"path": folder_path}
        self._post(_name=APINames.RSS, _method="addFolder", data=data, **kwargs)

    @Alias("rss_addFeed")
    @login_required
    def rss_add_feed(self, url=None, item_path=None, **kwargs):
        """
        Add new RSS feed. Folders in path must already exist. (alias: rss_addFeed)

        :raises Conflict409Error:

        :param url: URL of RSS feed (e.g http://thepiratebay.org/rss/top100/200)
        :param item_path: Name and/or path for new feed (e.g. Folder\Subfolder\FeedName)
        :return: None
        """
        data = {"url": url, "path": item_path}
        self._post(_name=APINames.RSS, _method="addFeed", data=data, **kwargs)

    @Alias("rss_removeItem")
    @login_required
    def rss_remove_item(self, item_path=None, **kwargs):
        """
        Remove a RSS item (folder, feed, etc). (alias: rss_removeItem)

        NOTE: Removing a folder also removes everything in it.

        :raises Conflict409Error:

        :param item_path: path to item to be removed (e.g. Folder\Subfolder\ItemName)
        :return: None
        """
        data = {"path": item_path}
        self._post(_name=APINames.RSS, _method="removeItem", data=data, **kwargs)

    @Alias("rss_moveItem")
    @login_required
    def rss_move_item(self, orig_item_path=None, new_item_path=None, **kwargs):
        """
        Move/rename a RSS item (folder, feed, etc). (alias: rss_moveItem)

        :raises Conflict409Error:

        :param orig_item_path: path to item to be removed (e.g. Folder\Subfolder\ItemName)
        :param new_item_path: path to item to be removed (e.g. Folder\Subfolder\ItemName)
        :return: None
        """
        data = {"itemPath": orig_item_path, "destPath": new_item_path}
        self._post(_name=APINames.RSS, _method="moveItem", data=data, **kwargs)

    @response_json(RSSitemsDictionary)
    @login_required
    def rss_items(self, include_feed_data=None, **kwargs):
        """
        Retrieve RSS items and optionally feed data.

        :param include_feed_data: True or false to include feed data
        :return: :class:`RSSitemsDictionary`
        """
        params = {"withData": include_feed_data}
        return self._get(_name=APINames.RSS, _method="items", params=params, **kwargs)

    @endpoint_introduced("2.2", "rss/refreshItem")
    @Alias("rss_refreshItem")
    @login_required
    def rss_refresh_item(self, item_path=None, **kwargs):
        """
        Trigger a refresh for a RSS item (alias: rss_refreshItem)

        :param item_path: path to item to be refreshed (e.g. Folder\Subfolder\ItemName)
        :return: None
        """
        # HACK: v4.1.7 and v4.1.8 both use api v2.2; however, refreshItem was introduced in v4.1.8
        if self._is_version_less_than("v4.1.7", self.app.version, False):
            data = {"itemPath": item_path}
            self._post(_name=APINames.RSS, _method="refreshItem", data=data, **kwargs)

    @endpoint_introduced("2.5.1", "rss/markAsRead")
    @Alias("rss_markAsRead")
    @login_required
    def rss_mark_as_read(self, item_path=None, article_id=None, **kwargs):
        """
        Mark RSS article as read. If article ID is not provider, the entire feed is marked as read. (alias: rss_markAsRead)

        :raises NotFound404Error:

        :param item_path: path to item to be refreshed (e.g. Folder\Subfolder\ItemName)
        :param article_id: article ID from rss_items()
        :return: None
        """
        data = {"itemPath": item_path, "articleId": article_id}
        self._post(_name=APINames.RSS, _method="markAsRead", data=data, **kwargs)

    @Alias("rss_setRule")
    @login_required
    def rss_set_rule(self, rule_name=None, rule_def=None, **kwargs):
        """
        Create a new RSS auto-downloading rule. (alias: rss_setRule)

        :param rule_name: name for new rule
        :param rule_def: dictionary with rule fields - https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#set-auto-downloading-rule
        :return: None
        """
        data = {"ruleName": rule_name, "ruleDef": dumps(rule_def)}
        self._post(_name=APINames.RSS, _method="setRule", data=data, **kwargs)

    @Alias("rss_renameRule")
    @login_required
    def rss_rename_rule(self, orig_rule_name=None, new_rule_name=None, **kwargs):
        """
        Rename a RSS auto-download rule. (alias: rss_renameRule)
        Note: this endpoint did not work properly until qBittorrent v4.3.0

        :param orig_rule_name: current name of rule
        :param new_rule_name: new name for rule
        :return: None
        """
        data = {"ruleName": orig_rule_name, "newRuleName": new_rule_name}
        self._post(_name=APINames.RSS, _method="renameRule", data=data, **kwargs)

    @Alias("rss_removeRule")
    @login_required
    def rss_remove_rule(self, rule_name=None, **kwargs):
        """
        Delete a RSS auto-downloading rule. (alias: rss_removeRule)

        :param rule_name: Name of rule to delete
        :return: None
        """
        data = {"ruleName": rule_name}
        self._post(_name=APINames.RSS, _method="removeRule", data=data, **kwargs)

    @response_json(RSSRulesDictionary)
    @login_required
    def rss_rules(self, **kwargs):
        """
        Retrieve RSS auto-download rule definitions.

        :return: :class:`RSSRulesDictionary`
        """
        return self._get(_name=APINames.RSS, _method="rules", **kwargs)

    @endpoint_introduced("2.5.1", "rss/matchingArticles")
    @Alias("rss_matchingArticles")
    @response_json(RSSitemsDictionary)
    @login_required
    def rss_matching_articles(self, rule_name=None, **kwargs):
        """
        Fetch all articles matching a rule. (alias: rss_matchingArticles)

        :param rule_name: Name of rule to return matching articles
        :return: :class:`RSSitemsDictionary`
        """
        data = {"ruleName": rule_name}
        return self._post(
            _name=APINames.RSS, _method="matchingArticles", data=data, **kwargs
        )



from qbittorrentapi.client import Client
from qbittorrentapi.definitions import *
from qbittorrentapi.exceptions import *

from qbittorrentapi.app import *
from qbittorrentapi.auth import *
from qbittorrentapi.log import *
from qbittorrentapi.request import *
from qbittorrentapi.rss import *
from qbittorrentapi.search import *
from qbittorrentapi.sync import *
from qbittorrentapi.torrents import *
from qbittorrentapi.transfer import *



from logging import getLogger
from logging import NullHandler
from os import environ
from pkg_resources import parse_version
from time import sleep

try:  # python 3
    from collections.abc import Iterable
    from urllib.parse import urljoin
    from urllib.parse import urlparse
except ImportError:  # python 2
    from collections import Iterable
    from urlparse import urljoin
    from urlparse import urlparse

from requests import exceptions as requests_exceptions
from requests import head as requests_head
from requests import Session
from requests.adapters import HTTPAdapter
from six import string_types as six_string_types
from urllib3 import disable_warnings
from urllib3.exceptions import InsecureRequestWarning
from urllib3.util.retry import Retry

from qbittorrentapi.definitions import APINames
from qbittorrentapi.exceptions import APIConnectionError
from qbittorrentapi.exceptions import HTTPError
from qbittorrentapi.exceptions import HTTP5XXError
from qbittorrentapi.exceptions import MissingRequiredParameters400Error
from qbittorrentapi.exceptions import InvalidRequest400Error
from qbittorrentapi.exceptions import Unauthorized401Error
from qbittorrentapi.exceptions import Forbidden403Error
from qbittorrentapi.exceptions import NotFound404Error
from qbittorrentapi.exceptions import Conflict409Error
from qbittorrentapi.exceptions import UnsupportedMediaType415Error
from qbittorrentapi.exceptions import InternalServerError500Error

logger = getLogger(__name__)
getLogger("qbittorrentapi").addHandler(NullHandler())


class HelpersMixIn(object):
    """
    Miscellaneous helper functions.
    """

    @classmethod
    def _list2string(cls, input_list=None, delimiter="|"):
        """
        Convert entries in a list to a concatenated string

        :param input_list: list to convert
        :param delimiter: delimiter for concatenation
        :return: if input is a list, concatenated string...else whatever the input was
        """
        if not isinstance(input_list, six_string_types) and isinstance(
            input_list, Iterable
        ):
            return delimiter.join(map(str, input_list))
        return input_list

    @classmethod
    def _suppress_context(cls, exc):
        """
        This is used to mask an exception with another one.

        For instance, below, the divide by zero error is masked by the CustomException.
            try:
                1/0
            except ZeroDivisionError:
                raise suppress_context(CustomException())

        Note: In python 3, the last line would simply be raise CustomException() from None
        :param exc: new Exception that will be raised
        :return: Exception to be raised
        """
        exc.__cause__ = None
        return exc

    @classmethod
    def _is_version_less_than(cls, ver1, ver2, lteq=True):
        """
        Determine if ver1 is equal to or later than ver2.

        Note: changes need to be reflected in decorators._is_version_less_than as well

        :param ver1: version to check
        :param ver2: current version of application
        :param lteq: True for Less Than or Equals; False for just Less Than
        :return: True or False
        """
        if lteq:
            return parse_version(ver1) <= parse_version(ver2)
        return parse_version(ver1) < parse_version(ver2)


class Request(HelpersMixIn):
    """Facilitates HTTP requests to qBittorrent."""

    def __init__(self, host="", port=None, username=None, password=None, **kwargs):
        self.host = host
        self.port = port
        self.username = username or ""
        self._password = password or ""

        self._initialize_context()
        self._initialize_lesser(**kwargs)

        # turn off console-printed warnings about SSL certificate issues.
        # these errors are only shown once the user has explicitly allowed
        # untrusted certs via VERIFY_WEBUI_CERTIFICATE...so printing them
        # in a console isn't particularly useful.
        if not self._VERIFY_WEBUI_CERTIFICATE:
            disable_warnings(InsecureRequestWarning)

    def _initialize_context(self):
        """
        Initialize and/or reset communications context with qBittorrent.
        This is necessary on startup or when the auth cookie needs to be replaced...perhaps
        because it expired, qBittorrent was restarted, significant settings changes, etc.
        """
        # base path for all API endpoints
        self._API_BASE_PATH = "api/v2"

        # reset URL so the full URL is derived again (primarily allows for switching scheme for WebUI: HTTP <-> HTTPS)
        self._API_BASE_URL = None

        # reset Requests session so it is rebuilt with new auth cookie and all
        self._requests_session = None

        # reinitialize interaction layers
        self._application = None
        self._authorization = None
        self._transfer = None
        self._torrents = None
        self._torrent_categories = None
        self._torrent_tags = None
        self._log = None
        self._sync = None
        self._rss = None
        self._search = None

    def _initialize_lesser(
        self,
        EXTRA_HEADERS=None,
        VERIFY_WEBUI_CERTIFICATE=True,
        RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=False,
        RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS=False,
        VERBOSE_RESPONSE_LOGGING=False,
        PRINT_STACK_FOR_EACH_REQUEST=False,
        SIMPLE_RESPONSES=False,
        DISABLE_LOGGING_DEBUG_OUTPUT=False,
        MOCK_WEB_API_VERSION=None,
    ):
        """Initialize lessor used configuration"""

        # Configuration parameters
        self._EXTRA_HEADERS = EXTRA_HEADERS if isinstance(EXTRA_HEADERS, dict) else {}
        self._VERIFY_WEBUI_CERTIFICATE = bool(VERIFY_WEBUI_CERTIFICATE)
        self._VERBOSE_RESPONSE_LOGGING = bool(VERBOSE_RESPONSE_LOGGING)
        self._PRINT_STACK_FOR_EACH_REQUEST = bool(PRINT_STACK_FOR_EACH_REQUEST)
        self._SIMPLE_RESPONSES = bool(SIMPLE_RESPONSES)
        self._RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS = bool(
            RAISE_UNIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS
            or RAISE_NOTIMPLEMENTEDERROR_FOR_UNIMPLEMENTED_API_ENDPOINTS
        )
        if bool(DISABLE_LOGGING_DEBUG_OUTPUT):
            for logger_ in ("qbittorrentapi", "requests", "urllib3"):
                if getLogger(logger_).level < 20:
                    getLogger(logger_).setLevel("INFO")

        # Environment variables have lowest priority
        if self.host == "" and environ.get("PYTHON_QBITTORRENTAPI_HOST") is not None:
            logger.debug(
                "Using PYTHON_QBITTORRENTAPI_HOST env variable for qBittorrent hostname"
            )
            self.host = environ["PYTHON_QBITTORRENTAPI_HOST"]
        if (
            self.username == ""
            and environ.get("PYTHON_QBITTORRENTAPI_USERNAME") is not None
        ):
            logger.debug(
                "Using PYTHON_QBITTORRENTAPI_USERNAME env variable for username"
            )
            self.username = environ["PYTHON_QBITTORRENTAPI_USERNAME"]
        if (
            self._password == ""
            and environ.get("PYTHON_QBITTORRENTAPI_PASSWORD") is not None
        ):
            logger.debug(
                "Using PYTHON_QBITTORRENTAPI_PASSWORD env variable for password"
            )
            self._password = environ["PYTHON_QBITTORRENTAPI_PASSWORD"]
        if (
            self._VERIFY_WEBUI_CERTIFICATE is True
            and environ.get("PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE")
            is not None
        ):
            self._VERIFY_WEBUI_CERTIFICATE = False

        # Mocking variables until better unit testing exists
        self._MOCK_WEB_API_VERSION = MOCK_WEB_API_VERSION

    def _get(self, _name=APINames.EMPTY, _method="", **kwargs):
        return self._request_manager(
            http_method="get", api_namespace=_name, api_method=_method, **kwargs
        )

    def _post(self, _name=APINames.EMPTY, _method="", **kwargs):
        return self._request_manager(
            http_method="post", api_namespace=_name, api_method=_method, **kwargs
        )

    def _request_manager(self, _retries=2, _retry_backoff_factor=0.3, **kwargs):
        """
        Wrapper to manage request retries and severe exceptions.

        This should retry at least twice to account for the Web API switching from HTTP to HTTPS.
        During the second attempt, the URL is rebuilt using HTTP or HTTPS as appropriate.
        """

        def build_error_msg(exc):
            """Create error message for exception to be raised to user."""
            error_prologue = "Failed to connect to qBittorrent. "
            error_messages = {
                requests_exceptions.SSLError: "This is likely due to using an untrusted certificate "
                "(likely self-signed) for HTTPS qBittorrent WebUI. To suppress this error (and skip "
                "certificate verification consequently exposing the HTTPS connection to man-in-the-middle "
                "attacks), set VERIFY_WEBUI_CERTIFICATE=False when instantiating Client or set "
                "environment variable PYTHON_QBITTORRENTAPI_DO_NOT_VERIFY_WEBUI_CERTIFICATE "
                "to a non-null value. SSL Error: %s" % repr(exc),
                requests_exceptions.HTTPError: "Invalid HTTP Response: %s" % repr(exc),
                requests_exceptions.TooManyRedirects: "Too many redirects: %s"
                % repr(exc),
                requests_exceptions.ConnectionError: "Connection Error: %s" % repr(exc),
                requests_exceptions.Timeout: "Timeout Error: %s" % repr(exc),
                requests_exceptions.RequestException: "Requests Error: %s" % repr(exc),
            }
            err_msg = error_messages.get(type(exc), "Unknown Error: %s" % repr(exc))
            err_msg = error_prologue + err_msg
            logger.debug(err_msg)
            return err_msg

        def retry_backoff(retry_count):
            """Back off on attempting each subsequent request retry."""
            if retry_count > 0:
                # The first retry is always immediate. if the backoff factor is 0.3,
                # then will sleep for 0s then .3s, then .6s, etc. between retries.
                backoff_time = _retry_backoff_factor * (2 ** ((retry_count + 1) - 1))
                sleep(backoff_time if backoff_time <= 10 else 10)
            logger.debug("Retry attempt %d" % (retry_count + 1))

        max_retries = _retries if _retries > 1 else 2
        for retry in range(0, (max_retries + 1)):
            try:
                return self._request(**kwargs)
            except HTTPError as e:
                # retry the request for HTTP 500 statuses;
                # raise immediately for other HTTP errors (e.g. 4XX statuses)
                if not isinstance(e, HTTP5XXError) or retry >= max_retries:
                    raise
            except Exception as e:
                if retry >= max_retries:
                    error_message = build_error_msg(exc=e)
                    response = getattr(e, "response", None)
                    raise APIConnectionError(error_message, response=response)

            retry_backoff(retry_count=retry)
            self._initialize_context()

    def _request(self, http_method, api_namespace, api_method, **kwargs):
        """
        Meat and potatoes of sending requests to qBittorrent.

        :param http_method: 'get' or 'post'
        :param api_namespace: the namespace for the API endpoint (e.g. torrents)
        :param api_method: the namespace for the API endpoint (e.g. torrents)
        :param kwargs: see _normalize_requests_params for additional support
        :return: Requests response
        """
        url = self._build_url(api_namespace=api_namespace, api_method=api_method)
        kwargs = self._trim_known_kwargs(kwargs=kwargs)
        params = self._normalize_requests_params(http_method=http_method, **kwargs)

        response = self._session.request(method=http_method, url=url, **params)

        self.verbose_logging(http_method=http_method, response=response, url=url)
        self.handle_error_responses(params=params, response=response)
        return response

    @staticmethod
    def _trim_known_kwargs(kwargs):
        """
        Since any extra keyword arguments from the user are automatically
        included in the request to qBittorrent, this removes any "known"
        arguments that definitely shouldn't be sent to qBittorrent.
        Generally, these are removed in previous processing, but in
        certain circumstances, they can survive in to request.

        :param kwargs: extra keywords arguments to be passed along in request
        :return: sanitized arguments
        """
        kwargs.pop("SIMPLE_RESPONSES", None)
        kwargs.pop("SIMPLE_RESPONSE", None)
        return kwargs

    def _build_url(self, api_namespace, api_method):
        """
        Create a fully qualified URL for the API endpoint.

        :param api_namespace: the namespace for the API endpoint (e.g. torrents)
        :param api_method: the specific method for the API endpoint (e.g. info)
        :return: urllib URL object
        """
        self._API_BASE_URL = self._build_base_url(
            base_url=self._API_BASE_URL,
            host=self.host,
            port=self.port,
        )
        return self._build_url_path(
            base_url=self._API_BASE_URL,
            api_base_path=self._API_BASE_PATH,
            api_namespace=api_namespace,
            api_method=api_method,
        )

    @staticmethod
    def _build_base_url(base_url=None, host="", port=None):
        """
        Determine the Base URL for the Web API endpoints.

        A URL is only actually built here if it's the first time here or
        the context was re-initialized. Otherwise, the most recently
        built URL is used.

        If the user doesn't provide a scheme for the URL, it will try HTTP
        first and fall back to HTTPS if that doesn't work. While this is
        probably backwards, qBittorrent or an intervening proxy can simply
        redirect to HTTPS and that'll be respected.

        Additionally, if users want to augment the path to the API endpoints,
        any path provided here will be preserved in the returned Base URL
        and prefixed to all subsequent API calls.

        :param base_url: if the URL was already built, this is the base URL
        :param host: user provided hostname for WebUI
        :return: base URL for Web API endpoint
        """
        if base_url is not None:
            return base_url

        # urlparse requires some sort of schema for parsing to work at all
        if not host.lower().startswith(("http:", "https:", "//")):
            host = "//" + host
        base_url = urlparse(url=host)
        logger.debug("Parsed user URL: %s" % repr(base_url))
        # default to HTTP if user didn't specify
        user_scheme = base_url.scheme
        base_url = base_url._replace(scheme="http") if not user_scheme else base_url
        alt_scheme = "https" if base_url.scheme == "http" else "http"
        # add port number if host doesn't contain one
        if port is not None and not isinstance(base_url.port, int):
            base_url = base_url._replace(netloc="%s:%s" % (base_url.netloc, port))

        # detect whether Web API is configured for HTTP or HTTPS
        logger.debug("Detecting scheme for URL...")
        try:
            # skip verification here...if there's a problem, we'll catch it during the actual API call
            r = requests_head(base_url.geturl(), allow_redirects=True, verify=False)
            # if WebUI eventually supports sending a redirect from HTTP to HTTPS then
            # Requests will automatically provide a URL using HTTPS.
            # For instance, the URL returned below will use the HTTPS scheme.
            #  >>> requests.head('http://grc.com', allow_redirects=True).url
            scheme = urlparse(r.url).scheme
        except requests_exceptions.RequestException:
            # assume alternative scheme will work...we'll fail later if neither are working
            scheme = alt_scheme

        # use detected scheme
        logger.debug("Using %s scheme" % scheme.upper())
        base_url = base_url._replace(scheme=scheme)
        if user_scheme and user_scheme != scheme:
            logger.warning(
                "Using '%s' instead of requested '%s' to communicate with qBittorrent"
                % (scheme, user_scheme)
            )

        # ensure URL always ends with a forward-slash
        base_url = base_url.geturl()
        if not base_url.endswith("/"):
            base_url = base_url + "/"
        logger.debug("Base URL: %s" % base_url)

        return base_url

    @staticmethod
    def _build_url_path(base_url, api_base_path, api_namespace, api_method):
        """
        Determine the full URL path for the API endpoint.

        :param base_url: base URL for API (e.g. http://localhost:8080 or http://example.com/qbt/)
        :param api_base_path: qBittorrent defined API path prefix (i.e. api/v2/)
        :param api_namespace: the namespace for the API endpoint (e.g. torrents)
        :param api_method: the specific method for the API endpoint (e.g. info)
        :return: full urllib URL object for API endpoint
                 (e.g. http://localhost:8080/api/v2/torrents/info or http://example.com/qbt/api/v2/torrents/info)
        """

        def sanitize(piece):
            """Ensure each piece of api path is a string without leading or trailing slashes"""
            return str(piece or "").strip("/")

        if isinstance(api_namespace, APINames):
            api_namespace = api_namespace.value
        api_path = "/".join(map(sanitize, (api_base_path, api_namespace, api_method)))
        # since base_url is guaranteed to end in a slash and api_path will never
        # start with a slash, this join only ever append to the path in base_url
        url = urljoin(base_url, api_path)
        return url

    @property
    def _session(self):
        """
        Create or return existing Requests session.

        :return: Requests Session object
        """
        if self._requests_session:
            return self._requests_session

        self._requests_session = Session()

        # default headers to prevent qBittorrent throwing any alarms
        self._requests_session.headers.update(
            {"Referer": self._API_BASE_URL, "Origin": self._API_BASE_URL}
        )

        # add any user-defined headers to be sent in all requests
        self._requests_session.headers.update(self._EXTRA_HEADERS)

        # enable/disable TLS verification for all requests
        self._requests_session.verify = self._VERIFY_WEBUI_CERTIFICATE

        # enable retries in Requests if HTTP call fails.
        # this is sorta doubling up on retries since request_manager() will
        # attempt retries as well. however, these retries will not delay.
        # so, if the problem is just a network blip then Requests will
        # automatically retry (with zero delay) and probably fix the issue
        # without coming all the way back to requests_wrapper. if this retries
        # is increased much above 1, and backoff_factor is non-zero, this
        # will start adding noticeable delays in these retry attempts...which
        # would then compound with similar delay logic in request_manager.
        # at any rate, the retries count in request_manager should always be
        # at least 2 to accommodate significant settings changes in qBittorrent
        # such as enabling HTTPs in Web UI settings.
        retries = 1
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            status_forcelist={500, 502, 504},
            raise_on_status=False,
        )
        adapter = HTTPAdapter(max_retries=retry)
        self._requests_session.mount("http://", adapter)
        self._requests_session.mount("https://", adapter)

        return self._requests_session

    @staticmethod
    def _normalize_requests_params(
        http_method,
        headers=None,
        data=None,
        params=None,
        files=None,
        requests_params=None,
        **kwargs
    ):
        """
        Extract several keyword arguments to send in the request.

        :param headers: additional headers to send with the request
        :param data: key/value pairs to send as body
        :param params: key/value pairs to send as query parameters
        :param files: key/value pairs to include as multipart POST requests
        :param requests_params: keyword arguments for call to Requests
        :return: dictionary of parameters for Requests call
        """
        # these are completely user defined and intended to allow users
        # of this client to control the behavior of Requests
        requests_params = requests_params or {}

        # these are expected to be populated by this client as necessary for qBittorrent
        data = data or {}
        params = params or {}
        files = files or {}

        # these are user-defined headers to include with the request
        headers = headers or {}
        # send Content-Length as 0 for empty POSTs...Requests will not send Content-Length
        # if data is empty but qBittorrent may complain otherwise
        if http_method == "post" and not any(filter(None, data.values())):
            headers["Content-Length"] = "0"

        # any other keyword arguments are sent to qBittorrent as part of the request.
        # These are user-defined since this Client will put everything in data/params/files
        # that needs to be sent to qBittorrent.
        if kwargs:
            if http_method == "get":
                params.update(kwargs)
            if http_method == "post":
                data.update(kwargs)

        d = dict(data=data, params=params, files=files, headers=headers)
        d.update(requests_params)
        return d

    @staticmethod
    def handle_error_responses(params, response):
        """Raise proper exception if qBittorrent returns Error HTTP Status."""
        if response.status_code < 400:
            # short circuit for non-error statuses
            return
        elif response.status_code == 400:
            # Returned for malformed requests such as missing or invalid parameters.
            #
            # If an error_message isn't returned, qBittorrent didn't receive all required parameters.
            # APIErrorType::BadParams
            # the name (i.e. Bad Response) of the HTTP error started being returned in v4.3.0
            if response.text in ("", "Bad Request"):
                raise MissingRequiredParameters400Error()
            raise InvalidRequest400Error(response.text)

        elif response.status_code == 401:
            # Primarily reserved for XSS and host header issues. Is also
            raise Unauthorized401Error(response.text)

        elif response.status_code == 403:
            # Not logged in or calling an API method that isn't public
            # APIErrorType::AccessDenied
            raise Forbidden403Error(response.text)

        elif response.status_code == 404:
            # API method doesn't exist or more likely, torrent not found
            # APIErrorType::NotFound
            error_message = response.text
            if error_message in ("", "Not Found"):
                error_torrent_hash = ""
                if params["data"]:
                    error_torrent_hash = params["data"].get("hash", error_torrent_hash)
                    error_torrent_hash = params["data"].get(
                        "hashes", error_torrent_hash
                    )
                if params and error_torrent_hash == "":
                    error_torrent_hash = params["params"].get(
                        "hash", error_torrent_hash
                    )
                    error_torrent_hash = params["params"].get(
                        "hashes", error_torrent_hash
                    )
                if error_torrent_hash:
                    error_message = "Torrent hash(es): %s" % error_torrent_hash
            raise NotFound404Error(error_message)

        elif response.status_code == 409:
            # APIErrorType::Conflict
            raise Conflict409Error(response.text)

        elif response.status_code == 415:
            # APIErrorType::BadData
            raise UnsupportedMediaType415Error(response.text)

        elif response.status_code >= 500:
            raise InternalServerError500Error(response.text)

        elif response.status_code >= 400:
            # Unaccounted for errors from API
            raise HTTPError(response.text)

    def verbose_logging(self, http_method, response, url):
        """Log verbose information about request. Can be useful during development."""
        if self._VERBOSE_RESPONSE_LOGGING:
            resp_logger = logger.debug
            max_text_length_to_log = 254
            if response.status_code != 200:
                # log as much as possible in a error condition
                max_text_length_to_log = 10000

            resp_logger("Request URL: (%s) %s" % (http_method.upper(), response.url))
            resp_logger("Request Headers: %s" % response.request.headers)
            if (
                str(response.request.body) not in ("None", "")
                and "auth/login" not in url
            ):
                body_len = (
                    max_text_length_to_log
                    if len(response.request.body) > max_text_length_to_log
                    else len(response.request.body)
                )
                resp_logger(
                    "Request body: %s%s"
                    % (
                        response.request.body[:body_len],
                        "...<truncated>" if body_len >= 200 else "",
                    )
                )

            resp_logger(
                "Response status: %s (%s)" % (response.status_code, response.reason)
            )
            if response.text:
                text_len = (
                    max_text_length_to_log
                    if len(response.text) > max_text_length_to_log
                    else len(response.text)
                )
                resp_logger(
                    "Response text: %s%s"
                    % (
                        response.text[:text_len],
                        "...<truncated>" if text_len >= 80 else "",
                    )
                )
        if self._PRINT_STACK_FOR_EACH_REQUEST:
            from traceback import print_stack

            print_stack()



from enum import Enum

try:
    from collections import UserList
except ImportError:
    from UserList import UserList

from qbittorrentapi._attrdict import AttrDict


class APINames(Enum):
    """
    API namespaces for API endpoints

    e.g 'torrents' in http://localhost:8080/api/v2/torrents/addTrackers
    """

    Authorization = "auth"
    Application = "app"
    Log = "log"
    Sync = "sync"
    Transfer = "transfer"
    Torrents = "torrents"
    RSS = "rss"
    Search = "search"
    EMPTY = ""


class TorrentStates(Enum):
    """
    Torrent States as defined by qBittorrent.

    Definitions:
        - wiki: https://github.com/qbittorrent/qBittorrent/wiki/WebUI-API-(qBittorrent-4.1)#get-torrent-list
        - code: https://github.com/qbittorrent/qBittorrent/blob/master/src/base/bittorrent/torrenthandle.h#L52

    :Usage:
        >>> from qbittorrentapi import Client
        >>> from qbittorrentapi import TorrentStates
        >>> client = Client()
        >>> # print torrent hashes for torrents that are downloading
        >>> for torrent in client.torrents_info():
        >>>     # check if torrent is downloading
        >>>     if torrent.state_enum.is_downloading:
        >>>         print(f'{torrent.hash} is downloading...')
        >>>     # the appropriate enum member can be directly derived
        >>>     state_enum = TorrentStates(torrent.state)
        >>>     print(f'{torrent.hash}: {state_enum.value}')
    """

    ERROR = "error"
    MISSING_FILES = "missingFiles"
    UPLOADING = "uploading"
    PAUSED_UPLOAD = "pausedUP"
    QUEUED_UPLOAD = "queuedUP"
    STALLED_UPLOAD = "stalledUP"
    CHECKING_UPLOAD = "checkingUP"
    FORCED_UPLOAD = "forcedUP"
    ALLOCATING = "allocating"
    DOWNLOADING = "downloading"
    METADATA_DOWNLOAD = "metaDL"
    PAUSED_DOWNLOAD = "pausedDL"
    QUEUED_DOWNLOAD = "queuedDL"
    FORCED_DOWNLOAD = "forcedDL"
    STALLED_DOWNLOAD = "stalledDL"
    CHECKING_DOWNLOAD = "checkingDL"
    CHECKING_RESUME_DATA = "checkingResumeData"
    MOVING = "moving"
    UNKNOWN = "unknown"

    @property
    def is_downloading(self):
        """Returns True if the State is categorized as Downloading."""
        return self in (
            TorrentStates.DOWNLOADING,
            TorrentStates.METADATA_DOWNLOAD,
            TorrentStates.STALLED_DOWNLOAD,
            TorrentStates.CHECKING_DOWNLOAD,
            TorrentStates.PAUSED_DOWNLOAD,
            TorrentStates.QUEUED_DOWNLOAD,
            TorrentStates.FORCED_DOWNLOAD,
        )

    @property
    def is_uploading(self):
        """Returns True if the State is categorized as Uploading."""
        return self in (
            TorrentStates.UPLOADING,
            TorrentStates.STALLED_UPLOAD,
            TorrentStates.CHECKING_UPLOAD,
            TorrentStates.QUEUED_UPLOAD,
            TorrentStates.FORCED_UPLOAD,
        )

    @property
    def is_complete(self):
        """Returns True if the State is categorized as Complete."""
        return self in (
            TorrentStates.UPLOADING,
            TorrentStates.STALLED_UPLOAD,
            TorrentStates.CHECKING_UPLOAD,
            TorrentStates.PAUSED_UPLOAD,
            TorrentStates.QUEUED_UPLOAD,
            TorrentStates.FORCED_UPLOAD,
        )

    @property
    def is_checking(self):
        """Returns True if the State is categorized as Checking."""
        return self in (
            TorrentStates.CHECKING_UPLOAD,
            TorrentStates.CHECKING_DOWNLOAD,
            TorrentStates.CHECKING_RESUME_DATA,
        )

    @property
    def is_errored(self):
        """Returns True if the State is categorized as Errored."""
        return self in (TorrentStates.MISSING_FILES, TorrentStates.ERROR)

    @property
    def is_paused(self):
        """Returns True if the State is categorized as Paused."""
        return self in (TorrentStates.PAUSED_UPLOAD, TorrentStates.PAUSED_DOWNLOAD)


class ClientCache(object):
    """Caches the client. Subclass this for any object that needs access to the Client."""

    def __init__(self, *args, **kwargs):
        self._client = kwargs.pop("client")
        super(ClientCache, self).__init__(*args, **kwargs)


class Dictionary(ClientCache, AttrDict):
    """Base definition of dictionary-like objects returned from qBittorrent."""

    def __init__(self, data=None, client=None):

        # iterate through a dictionary converting any nested dictionaries to AttrDicts
        def convert_dict_values_to_attrdicts(d):
            converted_dict = AttrDict()
            if isinstance(d, dict):
                for key, value in d.items():
                    # if the value is a dictionary, convert it to a AttrDict
                    if isinstance(value, dict):
                        # recursively send each value to convert its dictionary children
                        converted_dict[key] = convert_dict_values_to_attrdicts(
                            AttrDict(value)
                        )
                    else:
                        converted_dict[key] = value
                return converted_dict

        data = convert_dict_values_to_attrdicts(data)
        super(Dictionary, self).__init__(data or dict(), client=client)
        # allows updating properties that aren't necessarily a part of the AttrDict
        self._setattr("_allow_invalid_attributes", True)


class List(ClientCache, UserList):
    """Base definition for list-like objects returned from qBittorrent."""

    def __init__(self, list_entries=None, entry_class=None, client=None):

        entries = []
        for entry in list_entries:
            if isinstance(entry, dict):
                entries.append(entry_class(data=entry, client=client))
            else:
                entries.append(entry)
        super(List, self).__init__(entries, client=client)


class ListEntry(Dictionary):
    """Base definition for objects within a list returned from qBittorrent."""



# Copyright (c) 2013 Brendan Curran-Johnson
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Functions and Classes from attrdict.

*rmartin Feb 2021
AttrDict finally broke with Python 3.10 since abstract base classes can no
longer be imported from collections but should use collections.abc instead.
Since AttrDict is abandoned, I've consolidated the code here for future use.
AttrMap and AttrDefault are left for posterity but commented out.
"""

from abc import ABCMeta
from abc import abstractmethod
from re import match as re_match

try:  # python 3
    from collections.abc import Mapping
    from collections.abc import MutableMapping
    from collections.abc import Sequence
except ImportError:  # python 2
    from collections import Mapping
    from collections import MutableMapping
    from collections import Sequence

from six import add_metaclass as six_add_metaclass
from six import string_types as six_string_types
from six import binary_type as six_binary_type
from six import u as six_u


def merge(left, right):
    """
    Merge two mappings objects together, combining overlapping Mappings,
    and favoring right-values

    left: The left Mapping object.
    right: The right (favored) Mapping object.

    NOTE: This is not commutative (merge(a,b) != merge(b,a)).
    """
    merged = {}

    left_keys = frozenset(left)
    right_keys = frozenset(right)

    # Items only in the left Mapping
    for key in left_keys - right_keys:
        merged[key] = left[key]

    # Items only in the right Mapping
    for key in right_keys - left_keys:
        merged[key] = right[key]

    # in both
    for key in left_keys & right_keys:
        left_value = left[key]
        right_value = right[key]

        if isinstance(left_value, Mapping) and isinstance(
            right_value, Mapping
        ):  # recursive merge
            merged[key] = merge(left_value, right_value)
        else:  # overwrite with right value
            merged[key] = right_value

    return merged


@six_add_metaclass(ABCMeta)
class Attr(Mapping):
    """
    A mixin class for a mapping that allows for attribute-style access
    of values.

    A key may be used as an attribute if:
     * It is a string
     * It matches /^[A-Za-z][A-Za-z0-9_]*$/ (i.e., a public attribute)
     * The key doesn't overlap with any class attributes (for Attr,
        those would be 'get', 'items', 'keys', 'values', 'mro', and
        'register').

    If a values which is accessed as an attribute is a Sequence-type
    (and is not a string/bytes), it will be converted to a
    _sequence_type with any mappings within it converted to Attrs.

    NOTE: This means that if _sequence_type is not None, then a
        sequence accessed as an attribute will be a different object
        than if accessed as an attribute than if it is accessed as an
        item.
    """

    @abstractmethod
    def _configuration(self):
        """
        All required state for building a new instance with the same
        settings as the current object.
        """

    @classmethod
    def _constructor(cls, mapping, configuration):
        """
        A standardized constructor used internally by Attr.

        mapping: A mapping of key-value pairs. It is HIGHLY recommended
            that you use this as the internal key-value pair mapping, as
            that will allow nested assignment (e.g., attr.foo.bar = baz)
        configuration: The return value of Attr._configuration
        """
        raise NotImplementedError("You need to implement this")

    def __call__(self, key):
        """
        Dynamically access a key-value pair.

        key: A key associated with a value in the mapping.

        This differs from __getitem__, because it returns a new instance
        of an Attr (if the value is a Mapping object).
        """
        if key not in self:
            raise AttributeError(
                "'{cls} instance has no attribute '{name}'".format(
                    cls=self.__class__.__name__, name=key
                )
            )

        return self._build(self[key])

    def __getattr__(self, key):
        """
        Access an item as an attribute.
        """
        if key not in self or not self._valid_name(key):
            raise AttributeError(
                "'{cls}' instance has no attribute '{name}'".format(
                    cls=self.__class__.__name__, name=key
                )
            )

        return self._build(self[key])

    def __add__(self, other):
        """
        Add a mapping to this Attr, creating a new, merged Attr.

        other: A mapping.

        NOTE: Addition is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        return self._constructor(merge(self, other), self._configuration())

    def __radd__(self, other):
        """
        Add this Attr to a mapping, creating a new, merged Attr.

        other: A mapping.

        NOTE: Addition is not commutative. a + b != b + a.
        """
        if not isinstance(other, Mapping):
            return NotImplemented

        return self._constructor(merge(other, self), self._configuration())

    def _build(self, obj):
        """
        Conditionally convert an object to allow for recursive mapping
        access.

        obj: An object that was a key-value pair in the mapping. If obj
            is a mapping, self._constructor(obj, self._configuration())
            will be called. If obj is a non-string/bytes sequence, and
            self._sequence_type is not None, the obj will be converted
            to type _sequence_type and build will be called on its
            elements.
        """
        if isinstance(obj, Mapping):
            obj = self._constructor(obj, self._configuration())
        elif isinstance(obj, Sequence) and not isinstance(
            obj, (six_string_types, six_binary_type)
        ):
            sequence_type = getattr(self, "_sequence_type", None)

            if sequence_type:
                obj = sequence_type(self._build(element) for element in obj)

        return obj

    @classmethod
    def _valid_name(cls, key):
        """
        Check whether a key is a valid attribute name.

        A key may be used as an attribute if:
         * It is a string
         * It matches /^[A-Za-z][A-Za-z0-9_]*$/ (i.e., a public attribute)
         * The key doesn't overlap with any class attributes (for Attr,
            those would be 'get', 'items', 'keys', 'values', 'mro', and
            'register').
        """
        return (
            isinstance(key, six_string_types)
            and re_match("^[A-Za-z][A-Za-z0-9_]*$", key)
            and not hasattr(cls, key)
        )


@six_add_metaclass(ABCMeta)
class MutableAttr(Attr, MutableMapping):
    """
    A mixin class for a mapping that allows for attribute-style access
    of values.
    """

    def _setattr(self, key, value):
        """
        Add an attribute to the object, without attempting to add it as
        a key to the mapping.
        """
        super(MutableAttr, self).__setattr__(key, value)

    def __setattr__(self, key, value):
        """
        Add an attribute.

        key: The name of the attribute
        value: The attributes contents
        """
        if self._valid_name(key):
            self[key] = value
        elif getattr(self, "_allow_invalid_attributes", True):
            super(MutableAttr, self).__setattr__(key, value)
        else:
            raise TypeError(
                "'{cls}' does not allow attribute creation.".format(
                    cls=self.__class__.__name__
                )
            )

    def _delattr(self, key):
        """
        Delete an attribute from the object, without attempting to
        remove it from the mapping.
        """
        super(MutableAttr, self).__delattr__(key)

    def __delattr__(self, key, force=False):
        """
        Delete an attribute.

        key: The name of the attribute
        """
        if self._valid_name(key):
            del self[key]
        elif getattr(self, "_allow_invalid_attributes", True):
            super(MutableAttr, self).__delattr__(key)
        else:
            raise TypeError(
                "'{cls}' does not allow attribute deletion.".format(
                    cls=self.__class__.__name__
                )
            )


class AttrDict(dict, MutableAttr):
    """
    A dict that implements MutableAttr.
    """

    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)

        self._setattr("_sequence_type", tuple)
        self._setattr("_allow_invalid_attributes", False)

    def _configuration(self):
        """
        The configuration for an attrmap instance.
        """
        return self._sequence_type

    def __getstate__(self):
        """
        Serialize the object.
        """
        return (self.copy(), self._sequence_type, self._allow_invalid_attributes)

    def __setstate__(self, state):
        """
        Deserialize the object.
        """
        mapping, sequence_type, allow_invalid_attributes = state
        self.update(mapping)
        self._setattr("_sequence_type", sequence_type)
        self._setattr("_allow_invalid_attributes", allow_invalid_attributes)

    def __repr__(self):
        return six_u("AttrDict({contents})").format(
            contents=super(AttrDict, self).__repr__()
        )

    @classmethod
    def _constructor(cls, mapping, configuration):
        """
        A standardized constructor.
        """
        attr = cls(mapping)
        attr._setattr("_sequence_type", configuration)

        return attr


# class AttrMap(MutableAttr):
#     """
#     An implementation of MutableAttr.
#     """
#     def __init__(self, items=None, sequence_type=tuple):
#         if items is None:
#             items = {}
#         elif not isinstance(items, Mapping):
#             items = dict(items)
#
#         self._setattr('_sequence_type', sequence_type)
#         self._setattr('_mapping', items)
#         self._setattr('_allow_invalid_attributes', False)
#
#     def _configuration(self):
#         """
#         The configuration for an attrmap instance.
#         """
#         return self._sequence_type
#
#     def __getitem__(self, key):
#         """
#         Access a value associated with a key.
#         """
#         return self._mapping[key]
#
#     def __setitem__(self, key, value):
#         """
#         Add a key-value pair to the instance.
#         """
#         self._mapping[key] = value
#
#     def __delitem__(self, key):
#         """
#         Delete a key-value pair
#         """
#         del self._mapping[key]
#
#     def __len__(self):
#         """
#         Check the length of the mapping.
#         """
#         return len(self._mapping)
#
#     def __iter__(self):
#         """
#         Iterated through the keys.
#         """
#         return iter(self._mapping)
#
#     def __repr__(self):
#         """
#         Return a string representation of the object.
#         """
#         # sequence type seems like more trouble than it is worth.
#         # If people want full serialization, they can pickle, and in
#         # 99% of cases, sequence_type won't change anyway
#         return six_u("AttrMap({mapping})").format(mapping=repr(self._mapping))
#
#     def __getstate__(self):
#         """
#         Serialize the object.
#         """
#         return (
#             self._mapping,
#             self._sequence_type,
#             self._allow_invalid_attributes
#         )
#
#     def __setstate__(self, state):
#         """
#         Deserialize the object.
#         """
#         mapping, sequence_type, allow_invalid_attributes = state
#         self._setattr('_mapping', mapping)
#         self._setattr('_sequence_type', sequence_type)
#         self._setattr('_allow_invalid_attributes', allow_invalid_attributes)
#
#     @classmethod
#     def _constructor(cls, mapping, configuration):
#         """
#         A standardized constructor.
#         """
#         return cls(mapping, sequence_type=configuration)


# class AttrDefault(MutableAttr):
#     """
#     An implementation of MutableAttr with defaultdict support
#     """
#     def __init__(self, default_factory=None, items=None, sequence_type=tuple,
#                  pass_key=False):
#         if items is None:
#             items = {}
#         elif not isinstance(items, Mapping):
#             items = dict(items)
#
#         self._setattr('_default_factory', default_factory)
#         self._setattr('_mapping', items)
#         self._setattr('_sequence_type', sequence_type)
#         self._setattr('_pass_key', pass_key)
#         self._setattr('_allow_invalid_attributes', False)
#
#     def _configuration(self):
#         """
#         The configuration for a AttrDefault instance
#         """
#         return self._sequence_type, self._default_factory, self._pass_key
#
#     def __getitem__(self, key):
#         """
#         Access a value associated with a key.
#
#         Note: values returned will not be wrapped, even if recursive
#         is True.
#         """
#         if key in self._mapping:
#             return self._mapping[key]
#         elif self._default_factory is not None:
#             return self.__missing__(key)
#
#         raise KeyError(key)
#
#     def __setitem__(self, key, value):
#         """
#         Add a key-value pair to the instance.
#         """
#         self._mapping[key] = value
#
#     def __delitem__(self, key):
#         """
#         Delete a key-value pair
#         """
#         del self._mapping[key]
#
#     def __len__(self):
#         """
#         Check the length of the mapping.
#         """
#         return len(self._mapping)
#
#     def __iter__(self):
#         """
#         Iterated through the keys.
#         """
#         return iter(self._mapping)
#
#     def __missing__(self, key):
#         """
#         Add a missing element.
#         """
#         if self._pass_key:
#             self[key] = value = self._default_factory(key)
#         else:
#             self[key] = value = self._default_factory()
#
#         return value
#
#     def __repr__(self):
#         """
#         Return a string representation of the object.
#         """
#         return six_u(
#             "AttrDefault({default_factory}, {pass_key}, {mapping})"
#         ).format(
#             default_factory=repr(self._default_factory),
#             pass_key=repr(self._pass_key),
#             mapping=repr(self._mapping),
#         )
#
#     def __getstate__(self):
#         """
#         Serialize the object.
#         """
#         return (
#             self._default_factory,
#             self._mapping,
#             self._sequence_type,
#             self._pass_key,
#             self._allow_invalid_attributes,
#         )
#
#     def __setstate__(self, state):
#         """
#         Deserialize the object.
#         """
#         (default_factory, mapping, sequence_type, pass_key,
#          allow_invalid_attributes) = state
#
#         self._setattr('_default_factory', default_factory)
#         self._setattr('_mapping', mapping)
#         self._setattr('_sequence_type', sequence_type)
#         self._setattr('_pass_key', pass_key)
#         self._setattr('_allow_invalid_attributes', allow_invalid_attributes)
#
#     @classmethod
#     def _constructor(cls, mapping, configuration):
#         """
#         A standardized constructor.
#         """
#         sequence_type, default_factory, pass_key = configuration
#         return cls(default_factory, mapping, sequence_type=sequence_type,
#                    pass_key=pass_key)
